@@ -2,8 +2,8 @@
 #
 #  Pyomo: Python Optimization Modeling Objects
 #  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
-#  Under the terms of Contract DE-NA0003525 with National Technology and 
-#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
+#  Under the terms of Contract DE-NA0003525 with National Technology and
+#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
 #  rights in this software.
 #  This software is distributed under the 3-clause BSD License.
 #  ___________________________________________________________________________
@@ -19,25 +19,28 @@ __all__ = ('Objective',
 import sys
 import logging
 from weakref import ref as weakref_ref
-import inspect
 
 from pyomo.common.log import is_debug_set
-from pyomo.common.deprecation import deprecated
+from pyomo.common.deprecation import deprecated, RenamedClass
+from pyomo.common.formatting import tabular_writer
 from pyomo.common.timing import ConstructionTimer
 from pyomo.core.expr.numvalue import value
-from pyomo.core.base.plugin import ModelComponentFactory
-from pyomo.core.base.component import ActiveComponentData
-from pyomo.core.base.indexed_component import (ActiveIndexedComponent,
-                                               UnindexedComponent_set,
-                                               _get_indexed_component_data_name)
+from pyomo.core.base.component import (
+    ActiveComponentData, ModelComponentFactory,
+)
+from pyomo.core.base.indexed_component import (
+    ActiveIndexedComponent, UnindexedComponent_set, rule_wrapper,
+    _get_indexed_component_data_name,
+)
 from pyomo.core.base.expression import (_ExpressionData,
                                         _GeneralExpressionDataImpl)
-from pyomo.core.base.misc import apply_indexed_rule, tabular_writer
+from pyomo.core.base.misc import apply_indexed_rule
 from pyomo.core.base.set import Set
+from pyomo.core.base.initializer import (
+    Initializer, IndexedCallInitializer, CountedCallInitializer,
+)
 from pyomo.core.base import minimize, maximize
 
-from six import iteritems
-
 logger = logging.getLogger('pyomo.core')
 
 _rule_returned_none_error = """Objective '%s': rule returned None.
@@ -47,7 +50,7 @@ Objective.Skip.  The most common cause of this error is forgetting to
 include the "return" statement at the end of your rule.
 """
 
-def simple_objective_rule(fn):
+def simple_objective_rule(rule):
     """
     This is a decorator that translates None into Objective.Skip.
     This supports a simpler syntax in objective rules, though these
@@ -61,26 +64,10 @@ def simple_objective_rule(fn):
 
     model.o = Objective(rule=simple_objective_rule(...))
     """
+    return rule_wrapper(rule, {None: Objective.Skip})
 
-    def wrapper_function (*args, **kwargs):
-        #
-        # If the function is None, then skip this objective.
-        #
-        if fn is None:
-            return Objective.Skip
-        #
-        # Otherwise, the argument is a functor, so call it to generate
-        # the objective expression.
-        #
-        value = fn(*args, **kwargs)
-        if value is None:
-            return Objective.Skip
-        return value
-    return wrapper_function
-
-def simple_objectivelist_rule(fn):
+def simple_objectivelist_rule(rule):
     """
-
     This is a decorator that translates None into ObjectiveList.End.
     This supports a simpler syntax in objective rules, though these
     can be more difficult to debug when errors occur.
@@ -93,21 +80,7 @@ def simple_objectivelist_rule(fn):
 
     model.o = ObjectiveList(expr=simple_objectivelist_rule(...))
     """
-    def wrapper_function (*args, **kwargs):
-        #
-        # If the function is None, then the list is finished.
-        #
-        if fn is None:
-            return ObjectiveList.End
-        #
-        # Otherwise, the argument is a functor, so call it to generate
-        # the objective expression.
-        #
-        value = fn(*args, **kwargs)
-        if value is None:
-            return ObjectiveList.End
-        return value
-    return wrapper_function
+    return rule_wrapper(rule, {None: ObjectiveList.End})
 
 #
 # This class is a pure interface
@@ -171,8 +144,7 @@ class _GeneralObjectiveData(_GeneralExpressionDataImpl,
     """
 
     __pickle_slots__ = ("_sense",)
-    __slots__ = __pickle_slots__ + \
-                _GeneralExpressionDataImpl.__expression_slots__
+    __slots__ = __pickle_slots__ + _GeneralExpressionDataImpl.__pickle_slots__
 
     def __init__(self, expr=None, sense=minimize, component=None):
         _GeneralExpressionDataImpl.__init__(self, expr)
@@ -201,6 +173,11 @@ class _GeneralObjectiveData(_GeneralExpressionDataImpl,
     #       we don't need to implement a specialized __setstate__
     #       method.
 
+    def set_value(self, expr):
+        if expr is None:
+            raise ValueError(_rule_returned_none_error % (self.name,))
+        return super().set_value(expr)
+
     #
     # Abstract Interface
     #
@@ -216,8 +193,7 @@ class _GeneralObjectiveData(_GeneralExpressionDataImpl,
 
     def set_sense(self, sense):
         """Set the sense (direction) of this objective."""
-        if (sense == minimize) or \
-           (sense == maximize):
+        if sense in {minimize, maximize}:
             self._sense = sense
         else:
             raise ValueError("Objective sense must be set to one of "
@@ -233,165 +209,149 @@ class Objective(ActiveIndexedComponent):
     objectives to be used as part of expressions.
 
     Constructor arguments:
-        expr            
+        expr
             A Pyomo expression for this objective
-        rule            
+        rule
             A function that is used to construct objective expressions
-        sense           
+        sense
             Indicate whether minimizing (the default) or maximizing
-        doc             
+        doc
             A text string describing this component
-        name            
+        name
             A name for this component
 
     Public class attributes:
-        doc             
+        doc
             A text string describing this component
-        name            
+        name
             A name for this component
-        active          
-            A boolean that is true if this component will be used to construct 
+        active
+            A boolean that is true if this component will be used to construct
             a model instance
-        rule            
+        rule
             The rule used to initialize the objective(s)
-        sense           
+        sense
             The objective sense
 
     Private class attributes:
-        _constructed        
+        _constructed
             A boolean that is true if this component has been constructed
-        _data               
+        _data
             A dictionary from the index set to component data objects
-        _index              
+        _index
             The set of valid indices
-        _implicit_subsets   
+        _implicit_subsets
             A tuple of set objects that represents the index set
-        _model              
+        _model
             A weakref to the model that owns this component
-        _parent             
+        _parent
             A weakref to the parent block that owns this component
-        _type               
+        _type
             The class type for the derived subclass
     """
 
     _ComponentDataClass = _GeneralObjectiveData
-    NoObjective = (1000,)
-    Skip        = (1000,)
+    NoObjective = ActiveIndexedComponent.Skip
 
     def __new__(cls, *args, **kwds):
         if cls != Objective:
             return super(Objective, cls).__new__(cls)
         if not args or (args[0] is UnindexedComponent_set and len(args)==1):
-            return SimpleObjective.__new__(SimpleObjective)
+            return ScalarObjective.__new__(ScalarObjective)
         else:
             return IndexedObjective.__new__(IndexedObjective)
 
     def __init__(self, *args, **kwargs):
-        self._init_sense = kwargs.pop('sense', minimize)
-        self.rule  = kwargs.pop('rule', None)
-        self._init_expr  = kwargs.pop('expr', None)
+        _sense = kwargs.pop('sense', minimize)
+        _init = tuple( _arg for _arg in (
+            kwargs.pop('rule', None), kwargs.pop('expr', None)
+        ) if _arg is not None )
+        if len(_init) == 1:
+            _init = _init[0]
+        elif not _init:
+            _init = None
+        else:
+            raise ValueError("Duplicate initialization: Objective() only "
+                             "accepts one of 'rule=' and 'expr='")
+
         kwargs.setdefault('ctype', Objective)
         ActiveIndexedComponent.__init__(self, *args, **kwargs)
 
-    #
-    # TODO: Ideally we would not override these methods and instead add
-    # the contents of _check_skip_add to the set_value() method.
-    # Unfortunately, until IndexedComponentData objects know their own
-    # index, determining the index is a *very* expensive operation.  If
-    # we refactor things so that the Data objects have their own index,
-    # then we can remove these overloads.
-    #
-
-    def _setitem_impl(self, index, obj, value):
-        if self._check_skip_add(index, value) is None:
-            del self[index]
-            return None
-        else:
-            obj.set_value(value)
-            return obj
-
-    def _setitem_when_not_present(self, index, value):
-        if self._check_skip_add(index, value) is None:
-            return None
-        else:
-            return super(Objective, self)._setitem_when_not_present(
-                index=index, value=value)
+        self.rule = Initializer(_init)
+        self._init_sense = Initializer(_sense)
 
     def construct(self, data=None):
         """
         Construct the expression(s) for this objective.
         """
-        if is_debug_set(logger):
-            logger.debug(
-                "Constructing objective %s" % (self.name))
         if self._constructed:
             return
-        timer = ConstructionTimer(self)
         self._constructed = True
 
-        _init_expr = self._init_expr
-        _init_sense = self._init_sense
-        _init_rule = self.rule
-        #
-        # We no longer need these
-        #
-        self._init_expr = None
-        self._init_sense = None
-        # Utilities like DAE assume this stays around
-        #self.rule = None
+        timer = ConstructionTimer(self)
+        if is_debug_set(logger):
+            logger.debug("Constructing objective %s" % (self.name))
 
-        if (_init_rule is None) and \
-           (_init_expr is None):
-            # No construction rule or expression specified.
-            return
+        rule = self.rule
+        try:
+            # We do not (currently) accept data for constructing Objectives
+            index = None
+            assert data is None
 
-        _self_parent = self._parent()
-        if not self.is_indexed():
-            #
-            # Scalar component
-            #
-            if _init_rule is None:
-                tmp = _init_expr
-            else:
-                try:
-                    tmp = _init_rule(_self_parent)
-                except Exception:
-                    err = sys.exc_info()[1]
-                    logger.error(
-                        "Rule failed when generating expression for "
-                        "objective %s:\n%s: %s"
-                        % (self.name,
-                           type(err).__name__,
-                           err))
-                    raise
-            if self._setitem_when_not_present(None, tmp) is not None:
-                self.set_sense(_init_sense)
+            if rule is None:
+                # If there is no rule, then we are immediately done.
+                return
 
-        else:
-            if _init_expr is not None:
+            if rule.constant() and self.is_indexed():
                 raise IndexError(
-                    "Cannot initialize multiple indices of an "
-                    "objective with a single expression")
-            for ndx in self._index:
-                try:
-                    tmp = apply_indexed_rule(self,
-                                             _init_rule,
-                                             _self_parent,
-                                             ndx)
-                except Exception:
-                    err = sys.exc_info()[1]
-                    logger.error(
-                        "Rule failed when generating expression for"
-                        " objective %s with index %s:\n%s: %s"
-                        % (self.name,
-                           str(ndx),
-                           type(err).__name__,
-                           err))
-                    raise
-                ans = self._setitem_when_not_present(ndx, tmp)
-                if ans is not None:
-                    ans.set_sense(_init_sense)
-        timer.report()
+                    "Objective '%s': Cannot initialize multiple indices "
+                    "of an objective with a single expression" %
+                    (self.name,) )
+
+            block = self.parent_block()
+            if rule.contains_indices():
+                # The index is coming in externally; we need to validate it
+                for index in rule.indices():
+                    ans = self.__setitem__(index, rule(block, index))
+                    if ans is not None:
+                        self[index].set_sense(self._init_sense(block, index))
+            elif not self.index_set().isfinite():
+                # If the index is not finite, then we cannot iterate
+                # over it.  Since the rule doesn't provide explicit
+                # indices, then there is nothing we can do (the
+                # assumption is that the user will trigger specific
+                # indices to be created at a later time).
+                pass
+            else:
+                # Bypass the index validation and create the member directly
+                for index in self.index_set():
+                    ans = self._setitem_when_not_present(
+                        index, rule(block, index))
+                    if ans is not None:
+                        ans.set_sense(self._init_sense(block, index))
+        except Exception:
+            err = sys.exc_info()[1]
+            logger.error(
+                "Rule failed when generating expression for "
+                "Objective %s with index %s:\n%s: %s"
+                % (self.name,
+                   str(index),
+                   type(err).__name__,
+                   err))
+            raise
+        finally:
+            timer.report()
+
+    def _getitem_when_not_present(self, index):
+        if self.rule is None:
+            raise KeyError(index)
+        obj = self._setitem_when_not_present(
+            index, self.rule(self.parent_block(), index))
+        if obj is None:
+            raise KeyError(index)
+        else:
+            obj.set_sense(self._init_sense(block, index))
+        return obj
 
     def _pprint(self):
         """
@@ -402,7 +362,7 @@ class Objective(ActiveIndexedComponent):
              ("Index", self._index if self.is_indexed() else None),
              ("Active", self.active)
              ],
-            iteritems(self._data),
+            self._data.items(),
             ( "Active","Sense","Expression"),
             lambda k, v: [ v.active,
                            ("minimize" if (v.sense == minimize) else "maximize"),
@@ -426,39 +386,14 @@ class Objective(ActiveIndexedComponent):
 
         ostream.write("\n")
         tabular_writer( ostream, prefix+tab,
-                        ((k,v) for k,v in iteritems(self._data) if v.active),
+                        ((k,v) for k,v in self._data.items() if v.active),
                         ( "Active","Value" ),
                         lambda k, v: [ v.active, value(v), ] )
 
-    #
-    # Checks flags like Objective.Skip, etc. before
-    # actually creating an objective object. Optionally
-    # pass in the _ObjectiveData object to set the value
-    # on. Only returns the _ObjectiveData object when it
-    # should be added to the _data dict; otherwise, None
-    # is returned or an exception is raised.
-    #
-    def _check_skip_add(self, index, expr, objdata=None):
-        #
-        # Convert deprecated expression values
-        #
-        if expr is None:
-            raise ValueError(
-                _rule_returned_none_error %
-                (_get_indexed_component_data_name(self, index),) )
 
-        #
-        # Ignore an 'empty' objective
-        #
-        if expr.__class__ is tuple:
-            if expr == Objective.Skip:
-                return None
-
-        return expr
-
-class SimpleObjective(_GeneralObjectiveData, Objective):
+class ScalarObjective(_GeneralObjectiveData, Objective):
     """
-    SimpleObjective is the implementation representing a single,
+    ScalarObjective is the implementation representing a single,
     non-indexed objective.
     """
 
@@ -487,7 +422,7 @@ class SimpleObjective(_GeneralObjectiveData, Objective):
         if self._constructed:
             if len(self._data) == 0:
                 raise ValueError(
-                    "Accessing the expression of SimpleObjective "
+                    "Accessing the expression of ScalarObjective "
                     "'%s' before the Objective has been assigned "
                     "a sense or expression. There is currently "
                     "nothing to access." % (self.name))
@@ -504,13 +439,13 @@ class SimpleObjective(_GeneralObjectiveData, Objective):
 
     # for backwards compatibility reasons
     @property
-    @deprecated("The .value property getter on SimpleObjective is deprecated. "
+    @deprecated("The .value property getter on ScalarObjective is deprecated. "
                 "Use the .expr property getter instead", version='4.3.11323')
     def value(self):
         return self.expr
 
     @value.setter
-    @deprecated("The .value property setter on SimpleObjective is deprecated. "
+    @deprecated("The .value property setter on ScalarObjective is deprecated. "
                 "Use the set_value(expr) method instead", version='4.3.11323')
     def value(self, expr):
         self.set_value(expr)
@@ -521,7 +456,7 @@ class SimpleObjective(_GeneralObjectiveData, Objective):
         if self._constructed:
             if len(self._data) == 0:
                 raise ValueError(
-                    "Accessing the sense of SimpleObjective "
+                    "Accessing the sense of ScalarObjective "
                     "'%s' before the Objective has been assigned "
                     "a sense or expression. There is currently "
                     "nothing to access." % (self.name))
@@ -547,6 +482,9 @@ class SimpleObjective(_GeneralObjectiveData, Objective):
     # Objective.Skip but expects a valid expression or None
     #
 
+    def clear(self):
+        self._data = {}
+
     def set_value(self, expr):
         """Set the expression of this objective."""
         if not self._constructed:
@@ -555,13 +493,9 @@ class SimpleObjective(_GeneralObjectiveData, Objective):
                 "before the Objective has been constructed (there "
                 "is currently no object to set)."
                 % (self.name))
-
-        if len(self._data) == 0:
+        if not self._data:
             self._data[None] = self
-        if self._check_skip_add(None, expr) is None:
-            del self[None]
-            return None
-        return _GeneralObjectiveData.set_value(self, expr)
+        return super().set_value(expr)
 
     def set_sense(self, sense):
         """Set the sense (direction) of this objective."""
@@ -583,12 +517,18 @@ class SimpleObjective(_GeneralObjectiveData, Objective):
         """Add an expression with a given index."""
         if index is not None:
             raise ValueError(
-                "SimpleObjective object '%s' does not accept "
+                "ScalarObjective object '%s' does not accept "
                 "index values other than None. Invalid value: %s"
                 % (self.name, index))
         self.set_value(expr)
         return self
 
+
+class SimpleObjective(metaclass=RenamedClass):
+    __renamed__new_class__ = ScalarObjective
+    __renamed__version__ = '6.0'
+
+
 class IndexedObjective(Objective):
 
     #
@@ -611,74 +551,47 @@ class ObjectiveList(IndexedObjective):
     an index value is not specified.
     """
 
-    End             = (1003,)
+    class End(object): pass
 
     def __init__(self, **kwargs):
         """Constructor"""
-        args = (Set(dimen=1),)
         if 'expr' in kwargs:
             raise ValueError(
                 "ObjectiveList does not accept the 'expr' keyword")
-        Objective.__init__(self, *args, **kwargs)
+        _rule = kwargs.pop('rule', None)
+
+        args = (Set(dimen=1),)
+        super().__init__(*args, **kwargs)
+
+        self.rule = Initializer(_rule, allow_generators=True)
+        # HACK to make the "counted call" syntax work.  We wait until
+        # after the base class is set up so that is_indexed() is
+        # reliable.
+        if self.rule is not None and type(self.rule) is IndexedCallInitializer:
+            self.rule = CountedCallInitializer(self, self.rule)
 
     def construct(self, data=None):
         """
         Construct the expression(s) for this objective.
         """
-        generate_debug_messages = is_debug_set(logger)
-        if generate_debug_messages:
-            logger.debug(
-                "Constructing objective %s" % (self.name))
-
         if self._constructed:
             return
         self._constructed=True
+
+        if is_debug_set(logger):
+            logger.debug("Constructing objective list %s"
+                         % (self.name))
+
         self.index_set().construct()
 
-        assert self._init_expr is None
-        _init_rule = self.rule
-        _init_sense = self._init_sense
-
-        #
-        # We no longer need these
-        #
-        self._init_expr = None
-        self._init_sense = None
-        # Utilities like DAE assume this stays around
-        #self.rule = None
-
-
-        if _init_rule is None:
-            return
-
-        _generator = None
-        _self_parent = self._parent()
-        if inspect.isgeneratorfunction(_init_rule):
-            _generator = _init_rule(_self_parent)
-        elif inspect.isgenerator(_init_rule):
-            _generator = _init_rule
-        if _generator is None:
-            while True:
-                val = len(self._index) + 1
-                if generate_debug_messages:
-                    logger.debug(
-                        "   Constructing objective index "+str(val))
-                expr = apply_indexed_rule(self,
-                                          _init_rule,
-                                          _self_parent,
-                                          val)
-                if (expr.__class__ is tuple) and \
-                   (expr == ObjectiveList.End):
-                    return
-                self.add(expr, sense=_init_sense)
-
-        else:
-
-            for expr in _generator:
-                if (expr.__class__ is tuple) and \
-                   (expr == ObjectiveList.End):
-                    return
-                self.add(expr, sense=_init_sense)
+        if self.rule is not None:
+            _rule = self.rule(self.parent_block(), ())
+            for cc in iter(_rule):
+                if cc is ObjectiveList.End:
+                    break
+                if cc is Objective.Skip:
+                    continue
+                self.add(cc, sense=self._init_sense)
 
     def add(self, expr, sense=minimize):
         """Add an objective to the list."""
@@ -686,6 +599,8 @@ class ObjectiveList(IndexedObjective):
         self._index.add(next_idx)
         ans = self.__setitem__(next_idx, expr)
         if ans is not None:
+            if sense not in {minimize, maximize}:
+                sense = sense(self.parent_block(), next_idx)
             ans.set_sense(sense)
         return ans
 
